@@ -21,43 +21,22 @@
 
 QT_BEGIN_NAMESPACE
 
-Q_LOGGING_CATEGORY(lcAlsaOutput, "qt.multimedia.alsa.output")
+static Q_LOGGING_CATEGORY(lcAlsaOutput, "qt.multimedia.alsa.output")
 //#define DEBUG_AUDIO 1
 
 QAlsaAudioSink::QAlsaAudioSink(const QByteArray &device, QObject *parent)
     : QPlatformAudioSink(parent)
 {
-    bytesAvailable = 0;
-    handle = 0;
-    access = SND_PCM_ACCESS_RW_INTERLEAVED;
-    pcmformat = SND_PCM_FORMAT_S16;
-    buffer_frames = 0;
-    period_frames = 0;
-    buffer_size = 0;
-    period_size = 0;
-    buffer_time = 100000;
-    period_time = 20000;
-    totalTimeValue = 0;
-    audioBuffer = 0;
-    errorState = QAudio::NoError;
-    deviceState = QAudio::StoppedState;
-    audioSource = 0;
-    pullMode = true;
-    resuming = false;
-    opened = false;
-
-    m_volume = 1.0f;
-
     m_device = device;
 
     timer = new QTimer(this);
-    connect(timer,SIGNAL(timeout()),SLOT(userFeed()));
+    connect(timer, &QTimer::timeout, this, &QAlsaAudioSink::userFeed);
 }
 
 QAlsaAudioSink::~QAlsaAudioSink()
 {
     close();
-    disconnect(timer, SIGNAL(timeout()));
+    disconnect(timer, &QTimer::timeout, this, &QAlsaAudioSink::userFeed);
     QCoreApplication::processEvents();
     delete timer;
 }
@@ -136,21 +115,22 @@ int QAlsaAudioSink::setFormat()
         break;
     case QAudioFormat::Int16:
         if constexpr (QSysInfo::ByteOrder == QSysInfo::BigEndian)
-            pcmformat = SND_PCM_FORMAT_S16_LE;
-        else
             pcmformat = SND_PCM_FORMAT_S16_BE;
+        else
+            pcmformat = SND_PCM_FORMAT_S16_LE;
         break;
     case QAudioFormat::Int32:
         if constexpr (QSysInfo::ByteOrder == QSysInfo::BigEndian)
-            pcmformat = SND_PCM_FORMAT_S32_LE;
-        else
             pcmformat = SND_PCM_FORMAT_S32_BE;
+        else
+            pcmformat = SND_PCM_FORMAT_S32_LE;
         break;
     case QAudioFormat::Float:
         if constexpr (QSysInfo::ByteOrder == QSysInfo::BigEndian)
-            pcmformat = SND_PCM_FORMAT_FLOAT_LE;
-        else
             pcmformat = SND_PCM_FORMAT_FLOAT_BE;
+        else
+            pcmformat = SND_PCM_FORMAT_FLOAT_LE;
+        break;
     default:
         break;
     }
@@ -178,6 +158,11 @@ void QAlsaAudioSink::start(QIODevice* device)
     pullMode = true;
     audioSource = device;
 
+    connect(audioSource, &QIODevice::readyRead, timer, [this] {
+        if (!timer->isActive()) {
+            timer->start(period_time / 1000);
+        }
+    });
     deviceState = QAudio::ActiveState;
 
     open();
@@ -551,8 +536,7 @@ void QAlsaAudioSink::resume()
         }
         resuming = true;
 
-        deviceState = pullMode ? QAudio::ActiveState : QAudio::IdleState;
-
+        deviceState = suspendedInState;
         errorState = QAudio::NoError;
         timer->start(period_time/1000);
         emit stateChanged(deviceState);
@@ -572,6 +556,7 @@ QAudioFormat QAlsaAudioSink::format() const
 void QAlsaAudioSink::suspend()
 {
     if(deviceState == QAudio::ActiveState || deviceState == QAudio::IdleState || resuming) {
+        suspendedInState = deviceState;
         snd_pcm_drain(handle);
         timer->stop();
         deviceState = QAudio::SuspendedState;
@@ -627,11 +612,13 @@ bool QAlsaAudioSink::deviceReady()
 
         } else if(l == 0) {
             // Did not get any data to output
+            timer->stop();
+            snd_pcm_drain(handle);
             bytesAvailable = bytesFree();
             if(bytesAvailable > snd_pcm_frames_to_bytes(handle, buffer_frames-period_frames)) {
                 // Underrun
                 if (deviceState != QAudio::IdleState) {
-                    errorState = QAudio::UnderrunError;
+                    errorState = audioSource->atEnd() ? QAudio::NoError : QAudio::UnderrunError;
                     emit errorChanged(errorState);
                     deviceState = QAudio::IdleState;
                     emit stateChanged(deviceState);
