@@ -16,23 +16,29 @@
 
 #include <private/qtmultimediaglobal_p.h>
 #include <qmediarecorder.h>
+#include <qobject.h>
 
 QT_BEGIN_NAMESPACE
 
+class QCameraDevice;
 class QPlatformMediaIntegration;
 
-class Q_MULTIMEDIA_EXPORT QPlatformVideoDevices
+class Q_MULTIMEDIA_EXPORT QPlatformVideoDevices : public QObject
 {
+    Q_OBJECT
 public:
     QPlatformVideoDevices(QPlatformMediaIntegration *integration)
         : m_integration(integration)
     {}
-    virtual ~QPlatformVideoDevices();
 
-    virtual QList<QCameraDevice> videoDevices() const = 0;
+    ~QPlatformVideoDevices() override;
+
+    virtual QList<QCameraDevice> videoInputs() const = 0;
+
+Q_SIGNALS:
+    void videoInputsChanged();
 
 protected:
-    void videoInputsChanged();
     QPlatformMediaIntegration *m_integration = nullptr;
 };
 
