@@ -3,6 +3,7 @@
 
 #include "qvideoframeformat.h"
 #include "qvideotexturehelper_p.h"
+#include "qvideotransformation_p.h"
 
 #include <qdebug.h>
 #include <qlist.h>
@@ -12,7 +13,7 @@
 #include <qmatrix4x4.h>
 
 static void initResource() {
-    Q_INIT_RESOURCE(shaders);
+    Q_INIT_RESOURCE(qtmultimedia_shaders);
 }
 
 QT_BEGIN_NAMESPACE
@@ -33,13 +34,10 @@ public:
 
     bool operator ==(const QVideoFrameFormatPrivate &other) const
     {
-        if (pixelFormat == other.pixelFormat
-            && scanLineDirection == other.scanLineDirection
-            && frameSize == other.frameSize
-            && viewport == other.viewport
-            && frameRatesEqual(frameRate, other.frameRate)
-            && colorSpace == other.colorSpace
-            && mirrored == other.mirrored)
+        if (pixelFormat == other.pixelFormat && scanLineDirection == other.scanLineDirection
+            && frameSize == other.frameSize && viewport == other.viewport
+            && frameRatesEqual(frameRate, other.frameRate) && colorSpace == other.colorSpace
+            && transformation == other.transformation)
             return true;
 
         return false;
@@ -59,7 +57,7 @@ public:
     QRect viewport;
     float frameRate = 0.0;
     float maxLuminance = -1.;
-    bool mirrored = false;
+    VideoTransformation transformation;
 };
 
 QT_DEFINE_QESDP_SPECIALIZATION_DTOR(QVideoFrameFormatPrivate);
@@ -538,10 +536,28 @@ void QVideoFrameFormat::setScanLineDirection(Direction direction)
     d->scanLineDirection = direction;
 }
 
+#if QT_DEPRECATED_SINCE(6, 8)
 /*!
     Returns the frame rate of a video stream in frames per second.
 */
 qreal QVideoFrameFormat::frameRate() const
+{
+    return streamFrameRate();
+}
+
+/*!
+    Sets the frame \a rate of a video stream in frames per second.
+*/
+void QVideoFrameFormat::setFrameRate(qreal rate)
+{
+    setStreamFrameRate(rate);
+}
+#endif
+
+/*!
+    Returns the frame rate of a video stream in frames per second.
+*/
+qreal QVideoFrameFormat::streamFrameRate() const
 {
     return d->frameRate;
 }
@@ -549,7 +565,7 @@ qreal QVideoFrameFormat::frameRate() const
 /*!
     Sets the frame \a rate of a video stream in frames per second.
 */
-void QVideoFrameFormat::setFrameRate(qreal rate)
+void QVideoFrameFormat::setStreamFrameRate(qreal rate)
 {
     detach();
     d->frameRate = rate;
@@ -636,7 +652,13 @@ void QVideoFrameFormat::setColorRange(ColorRange range)
 
 /*!
     Returns \c true if the surface is mirrored around its vertical axis.
-    This is typically needed for video frames coming from a front camera of a mobile device.
+
+    Transformations of \c QVideoFrameFormat, specifically,
+    rotation and mirroring, can be determined by the orientation of
+    the camera sensor, camera settings, or the orientation of
+    the video stream.
+
+    Mirroring is applied after rotation.
 
     \note The mirroring here differs from QImage::mirrored, as a vertically mirrored QImage
     will be mirrored around its x-axis.
@@ -645,13 +667,20 @@ void QVideoFrameFormat::setColorRange(ColorRange range)
  */
 bool QVideoFrameFormat::isMirrored() const
 {
-    return d->mirrored;
+    return d->transformation.mirrorredHorizontallyAfterRotation;
 }
 
 /*!
     Sets if the surface is \a mirrored around its vertical axis.
-    This is typically needed for video frames coming from a front camera of a mobile device.
-    Default value is false.
+
+    Transformations of \c QVideoFrameFormat, specifically,
+    rotation and mirroring, can be determined by the orientation of
+    the camera sensor, camera settings, or the orientation of
+    the video stream.
+
+    Mirroring is applied after rotation.
+
+    Default value is \c false.
 
     \note The mirroring here differs from QImage::mirrored, as a vertically mirrored QImage
     will be mirrored around its x-axis.
@@ -661,7 +690,40 @@ bool QVideoFrameFormat::isMirrored() const
 void QVideoFrameFormat::setMirrored(bool mirrored)
 {
     detach();
-    d->mirrored = mirrored;
+    d->transformation.mirrorredHorizontallyAfterRotation = mirrored;
+}
+
+/*!
+    Returns the angle by which the surface is rotated clockwise.
+
+    Transformations of \c QVideoFrameFormat, specifically,
+    rotation and mirroring, can be determined by the orientation of
+    the camera sensor, camera settings, or the orientation of
+    the video stream.
+
+    Rotation is applied before mirroring.
+ */
+QtVideo::Rotation QVideoFrameFormat::rotation() const
+{
+    return d->transformation.rotation;
+}
+
+/*!
+    Sets the \a angle by which the surface is rotated clockwise.
+
+    Transformations of \c QVideoFrameFormat, specifically,
+    rotation and mirroring, can be determined by the orientation of
+    the camera sensor, camera settings, or the orientation of
+    the video stream.
+
+    Rotation is applied before mirroring.
+
+    Default value is \c QtVideo::Rotation::None.
+ */
+void QVideoFrameFormat::setRotation(QtVideo::Rotation angle)
+{
+    detach();
+    d->transformation.rotation = angle;
 }
 
 /*!
@@ -677,7 +739,7 @@ QString QVideoFrameFormat::vertexShaderFileName() const
 */
 QString QVideoFrameFormat::fragmentShaderFileName() const
 {
-    return QVideoTextureHelper::fragmentShaderFileName(*this);
+    return QVideoTextureHelper::fragmentShaderFileName(*this, nullptr);
 }
 
 /*!
@@ -743,7 +805,9 @@ QVideoFrameFormat::PixelFormat QVideoFrameFormat::pixelFormatFromImageFormat(QIm
     case QImage::Format_RGBA8888:
         return QVideoFrameFormat::Format_RGBA8888;
     case QImage::Format_RGBA8888_Premultiplied:
-        return QVideoFrameFormat::Format_ARGB8888_Premultiplied;
+        // QVideoFrameFormat::Format_RGBA8888_Premultiplied is to be added in 6.8
+        // Format_RGBX8888 suits the best as a workaround
+        return QVideoFrameFormat::Format_RGBX8888;
     case QImage::Format_RGBX8888:
         return QVideoFrameFormat::Format_RGBX8888;
     case QImage::Format_Grayscale8:
@@ -982,7 +1046,7 @@ QDebug operator<<(QDebug dbg, const QVideoFrameFormat &f)
         << "\n    frame size=" << f.frameSize()
         << "\n    viewport=" << f.viewport()
         << "\n    colorSpace=" << f.colorSpace()
-        << "\n    frameRate=" << f.frameRate()
+        << "\n    frameRate=" << f.streamFrameRate()
         << "\n    mirrored=" << f.isMirrored();
 
     return dbg;
