@@ -4,18 +4,20 @@
 #include "qquickvideooutput_p.h"
 
 #include <private/qvideooutputorientationhandler_p.h>
+#include <private/qvideoframetexturepool_p.h>
 #include <QtMultimedia/qmediaplayer.h>
 #include <QtMultimedia/qmediacapturesession.h>
 #include <private/qfactoryloader_p.h>
 #include <QtCore/qloggingcategory.h>
-#include <qvideosink.h>
 #include <QtQuick/QQuickWindow>
 #include <private/qquickwindow_p.h>
+#include <private/qmultimediautils_p.h>
 #include <qsgvideonode_p.h>
+#include <QtCore/qrunnable.h>
 
 QT_BEGIN_NAMESPACE
 
-Q_LOGGING_CATEGORY(qLcVideo, "qt.multimedia.video")
+static Q_LOGGING_CATEGORY(qLcVideo, "qt.multimedia.video")
 
 namespace {
 
@@ -24,23 +26,15 @@ inline bool qIsDefaultAspect(int o)
     return (o % 180) == 0;
 }
 
-/*
- * Return the orientation normalized to 0-359
- */
-inline int qNormalizedOrientation(int o)
+inline bool qIsDefaultAspect(QtVideo::Rotation rotation)
 {
-    // Negative orientations give negative results
-    int o2 = o % 360;
-    if (o2 < 0)
-        o2 += 360;
-    return o2;
+    return qIsDefaultAspect(qToUnderlying(rotation));
 }
-
 }
 
 /*!
     \qmltype VideoOutput
-    //! \instantiates QQuickVideoOutput
+    //! \nativetype QQuickVideoOutput
     \brief Render video or camera viewfinder.
 
     \ingroup multimedia_qml
@@ -103,16 +97,15 @@ QQuickVideoOutput::QQuickVideoOutput(QQuickItem *parent) :
 {
     setFlag(ItemHasContents, true);
 
-    m_sink = new QVideoSink(this);
+    m_sink = new QQuickVideoSink(this);
     qRegisterMetaType<QVideoFrameFormat>();
-    QObject::connect(m_sink, &QVideoSink::videoFrameChanged,
-                     [&](const QVideoFrame &frame) {
-                         setFrame(frame);
-                         emit frameUpdated(frame.size());
-                     });
 
-    QObject::connect(this, &QQuickVideoOutput::frameUpdated,
-                     this, &QQuickVideoOutput::_q_newFrame);
+    // TODO: investigate if we have any benefit of setting frame in the source thread
+    connect(m_sink, &QVideoSink::videoFrameChanged, this,
+            [this](const QVideoFrame &frame) {
+                setFrame(frame);
+            },
+            Qt::DirectConnection);
 
     initRhiForSink();
 }
@@ -171,9 +164,7 @@ void QQuickVideoOutput::_q_newFrame(QSize size)
 {
     update();
 
-    if (!qIsDefaultAspect(m_orientation + m_frameOrientation)) {
-        size.transpose();
-    }
+    size = qRotatedFrameSize(size, m_frameDisplayingRotation);
 
     if (m_nativeSize != size) {
         m_nativeSize = size;
@@ -225,18 +216,26 @@ void QQuickVideoOutput::_q_updateGeometry()
 /*!
     \qmlproperty int QtMultimedia::VideoOutput::orientation
 
-    In some cases the source video stream requires a certain
-    orientation to be correct.  This includes
+    This property determines the angle in, degrees, at which the displayed video
+    is rotated clockwise in video coordinates, where the Y-axis points
+    downwards on the display.
+
+    The orientation change affects the mapping of coordinates from the source to the viewport.
+
+    Only multiples of \c 90 degrees are supported, that is 0, 90, -90, 180, 270, etc.,
+    otherwise, the specified value is ignored.
+
+    In some cases, the source video stream requires a certain
+    orientation to be corrected. This includes
     sources like a camera viewfinder, where the displayed
-    viewfinder should match reality, no matter what rotation
+    viewfinder should match the reality, no matter what rotation
     the rest of the user interface has.
 
-    This property allows you to apply a rotation (in steps
-    of 90 degrees) to compensate for any user interface
-    rotation, with positive values in the anti-clockwise direction.
+    We recommend using this property to compensate a user interface
+    rotation, or align the output view with other application business
+    requirements.
 
-    The orientation change will also affect the mapping
-    of coordinates from source to viewport.
+    The default value is \c 0.
 */
 int QQuickVideoOutput::orientation() const
 {
@@ -255,7 +254,7 @@ void QQuickVideoOutput::setOrientation(int orientation)
 
     // If the new orientation is the same effect
     // as the old one, don't update the video node stuff
-    if ((m_orientation % 360) == (orientation % 360)) {
+    if (qVideoRotationFromDegrees(orientation - m_orientation) == QtVideo::Rotation::None) {
         m_orientation = orientation;
         emit orientationChanged();
         return;
@@ -270,6 +269,11 @@ void QQuickVideoOutput::setOrientation(int orientation)
 
     m_orientation = orientation;
 
+    {
+        QMutexLocker lock(&m_frameMutex);
+        m_frameDisplayingRotation = qNormalizedFrameTransformation(m_frame, m_orientation).rotation;
+    }
+
     if (oldAspect != newAspect) {
         m_nativeSize.transpose();
 
@@ -326,7 +330,7 @@ QRectF QQuickVideoOutput::sourceRect() const
     if (!size.isValid())
         return {};
 
-    if (!qIsDefaultAspect(m_orientation + m_frameOrientation))
+    if (!qIsDefaultAspect(m_frameDisplayingRotation))
         size.transpose();
 
 
@@ -354,7 +358,11 @@ void QQuickVideoOutput::geometryChange(const QRectF &newGeometry, const QRectF &
 
 void QQuickVideoOutput::_q_invalidateSceneGraph()
 {
-    invalidateSceneGraph();
+    // Invoked in the renderer thread
+
+    if (auto texturePool = m_texturePool.lock())
+        texturePool->clearTextures();
+    initRhiForSink();
 }
 
 void QQuickVideoOutput::_q_sceneGraphInitialized()
@@ -362,17 +370,17 @@ void QQuickVideoOutput::_q_sceneGraphInitialized()
     initRhiForSink();
 }
 
+void QQuickVideoOutput::_q_afterFrameEnd()
+{
+    if (auto texturePool = m_texturePool.lock())
+        texturePool->onFrameEndInvoked();
+}
+
 void QQuickVideoOutput::releaseResources()
 {
     // Called on the gui thread when the window is closed or changed.
-    invalidateSceneGraph();
-}
-
-void QQuickVideoOutput::invalidateSceneGraph()
-{
-    // Called on the render thread, e.g. when the context is lost.
-    //    QMutexLocker lock(&m_frameMutex);
     initRhiForSink();
+    QQuickItem::releaseResources();
 }
 
 void QQuickVideoOutput::initRhiForSink()
@@ -395,23 +403,25 @@ void QQuickVideoOutput::itemChange(QQuickItem::ItemChange change,
 
     if (m_window) {
         // We want to receive the signals in the render thread
-        QObject::connect(m_window, &QQuickWindow::sceneGraphInitialized, this, &QQuickVideoOutput::_q_sceneGraphInitialized,
-                         Qt::DirectConnection);
-        QObject::connect(m_window, &QQuickWindow::sceneGraphInvalidated,
-                         this, &QQuickVideoOutput::_q_invalidateSceneGraph, Qt::DirectConnection);
+        connect(m_window, &QQuickWindow::sceneGraphInitialized, this,
+                &QQuickVideoOutput::_q_sceneGraphInitialized, Qt::DirectConnection);
+        connect(m_window, &QQuickWindow::sceneGraphInvalidated, this,
+                &QQuickVideoOutput::_q_invalidateSceneGraph, Qt::DirectConnection);
+        connect(m_window, &QQuickWindow::afterFrameEnd, this, &QQuickVideoOutput::_q_afterFrameEnd,
+                Qt::DirectConnection);
     }
     initRhiForSink();
 }
 
 QSize QQuickVideoOutput::nativeSize() const
 {
-    return m_surfaceFormat.viewport().size();
+    return m_videoFormat.viewport().size();
 }
 
 void QQuickVideoOutput::updateGeometry()
 {
-    const QRectF viewport = m_surfaceFormat.viewport();
-    const QSizeF frameSize = m_surfaceFormat.frameSize();
+    const QRectF viewport = m_videoFormat.viewport();
+    const QSizeF frameSize = m_videoFormat.frameSize();
     const QRectF normalizedViewport(viewport.x() / frameSize.width(),
                                     viewport.y() / frameSize.height(),
                                     viewport.width() / frameSize.width(),
@@ -443,7 +453,7 @@ void QQuickVideoOutput::updateGeometry()
         const qreal totalWidth = normalizedViewport.width() * relativeWidth;
         const qreal totalHeight = normalizedViewport.height() * relativeHeight;
 
-        if (qIsDefaultAspect(orientation() + m_frameOrientation)) {
+        if (qIsDefaultAspect(m_frameDisplayingRotation)) {
             m_sourceTextureRect = QRectF(totalOffsetLeft, totalOffsetTop,
                                          totalWidth, totalHeight);
         } else {
@@ -451,18 +461,6 @@ void QQuickVideoOutput::updateGeometry()
                                          totalHeight, totalWidth);
         }
     }
-
-    if (m_surfaceFormat.scanLineDirection() == QVideoFrameFormat::BottomToTop) {
-        qreal top = m_sourceTextureRect.top();
-        m_sourceTextureRect.setTop(m_sourceTextureRect.bottom());
-        m_sourceTextureRect.setBottom(top);
-    }
-
-    if (m_surfaceFormat.isMirrored()) {
-        qreal left = m_sourceTextureRect.left();
-        m_sourceTextureRect.setLeft(m_sourceTextureRect.right());
-        m_sourceTextureRect.setRight(left);
-    }
 }
 
 QSGNode *QQuickVideoOutput::updatePaintNode(QSGNode *oldNode,
@@ -492,7 +490,9 @@ QSGNode *QQuickVideoOutput::updatePaintNode(QSGNode *oldNode,
             // Get a node that supports our frame. The surface is irrelevant, our
             // QSGVideoItemSurface supports (logically) anything.
             updateGeometry();
-            videoNode = new QSGVideoNode(this, m_surfaceFormat);
+            QRhi *rhi = m_window ? QQuickWindowPrivate::get(m_window)->rhi : nullptr;
+            videoNode = new QSGVideoNode(this, m_videoFormat, rhi);
+            m_texturePool = videoNode->texturePool();
             qCDebug(qLcVideo) << "updatePaintNode: Video node created. Handle type:" << m_frame.handleType();
         }
     }
@@ -506,37 +506,65 @@ QSGNode *QQuickVideoOutput::updatePaintNode(QSGNode *oldNode,
     if (m_frameChanged) {
         videoNode->setCurrentFrame(m_frame);
 
+        updateHdr(videoNode);
+
         //don't keep the frame for more than really necessary
         m_frameChanged = false;
         m_frame = QVideoFrame();
     }
 
-    // Negative rotations need lots of %360
-    videoNode->setTexturedRectGeometry(m_renderedRect, m_sourceTextureRect,
-                                       qNormalizedOrientation(orientation()));
+    videoNode->setTexturedRectGeometry(
+            m_renderedRect, m_sourceTextureRect,
+            VideoTransformation{ qVideoRotationFromDegrees(orientation()), m_mirrored });
 
     return videoNode;
 }
 
+void QQuickVideoOutput::updateHdr(QSGVideoNode *videoNode)
+{
+    auto *videoOutputWindow = window();
+    if (!videoOutputWindow)
+        return;
+
+    auto *swapChain = videoOutputWindow->swapChain();
+    if (!swapChain)
+        return;
+
+    const auto requiredSwapChainFormat = qGetRequiredSwapChainFormat(m_frame.surfaceFormat());
+    if (qShouldUpdateSwapChainFormat(swapChain, requiredSwapChainFormat)) {
+        auto *recreateSwapChainJob = QRunnable::create([swapChain, requiredSwapChainFormat]() {
+            swapChain->destroy();
+            swapChain->setFormat(requiredSwapChainFormat);
+            swapChain->createOrResize();
+        });
+
+        // Even though the 'recreate swap chain' job is scheduled for the current frame the
+        // effect will be visible only starting from the next frame since the recreation would
+        // happen after the actual swap.
+        videoOutputWindow->scheduleRenderJob(recreateSwapChainJob, QQuickWindow::AfterSwapStage);
+    }
+
+    videoNode->setSurfaceFormat(swapChain->format());
+    videoNode->setHdrInfo(swapChain->hdrInfo());
+}
+
 QRectF QQuickVideoOutput::adjustedViewport() const
 {
-    return m_surfaceFormat.viewport();
+    return m_videoFormat.viewport();
 }
 
 void QQuickVideoOutput::setFrame(const QVideoFrame &frame)
 {
-    m_frameMutex.lock();
-    m_surfaceFormat = frame.surfaceFormat();
-    m_frame = frame;
-    m_frameOrientation = frame.rotationAngle();
-    m_frameChanged = true;
-    m_frameMutex.unlock();
-}
+    {
+        QMutexLocker lock(&m_frameMutex);
 
-void QQuickVideoOutput::stop()
-{
-    setFrame({});
-    update();
+        m_videoFormat = frame.surfaceFormat();
+        m_frame = frame;
+        m_frameDisplayingRotation = qNormalizedFrameTransformation(frame, m_orientation).rotation;
+        m_frameChanged = true;
+    }
+
+    QMetaObject::invokeMethod(this, &QQuickVideoOutput::_q_newFrame, frame.size());
 }
 
 QT_END_NAMESPACE
