@@ -5,9 +5,11 @@
 #include <QtQuick/qsgmaterial.h>
 #include "qsgvideotexture_p.h"
 #include <QtMultimedia/private/qvideotexturehelper_p.h>
-#include <private/qquicktextnode_p.h>
+#include <private/qsginternaltextnode_p.h>
+#include <private/qquickitem_p.h>
 #include <private/qquickvideooutput_p.h>
-#include <private/qabstractvideobuffer_p.h>
+#include <private/qhwvideobuffer_p.h>
+#include <private/qvideoframetexturepool_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -41,15 +43,16 @@ class QSGVideoMaterialRhiShader : public QSGMaterialShader
 public:
     QSGVideoMaterialRhiShader(const QVideoFrameFormat &videoFormat,
                               const QRhiSwapChain::Format surfaceFormat,
-                              const QRhiSwapChainHdrInfo &hdrInfo)
+                              const QRhiSwapChainHdrInfo &hdrInfo,
+                              QRhi *rhi)
         : m_videoFormat(videoFormat)
         , m_surfaceFormat(surfaceFormat)
         , m_hdrInfo(hdrInfo)
     {
         setShaderFileName(VertexStage, QVideoTextureHelper::vertexShaderFileName(m_videoFormat));
-        setShaderFileName(
-                FragmentStage,
-                QVideoTextureHelper::fragmentShaderFileName(m_videoFormat, m_surfaceFormat));
+        setShaderFileName(FragmentStage,
+                          QVideoTextureHelper::fragmentShaderFileName(
+                                  m_videoFormat, rhi, m_surfaceFormat));
     }
 
     bool updateUniformData(RenderState &state, QSGMaterial *newMaterial,
@@ -67,15 +70,16 @@ protected:
 class QSGVideoMaterial : public QSGMaterial
 {
 public:
-    QSGVideoMaterial(const QVideoFrameFormat &videoFormat);
+    QSGVideoMaterial(const QVideoFrameFormat &videoFormat, QRhi *rhi);
 
     [[nodiscard]] QSGMaterialType *type() const override {
-        static QSGMaterialType type[QVideoFrameFormat::NPixelFormats];
-        return &type[m_videoFormat.pixelFormat()];
+        static constexpr int NFormats = QRhiSwapChain::HDRExtendedDisplayP3Linear + 1;
+        static QSGMaterialType type[QVideoFrameFormat::NPixelFormats][NFormats];
+        return &type[m_videoFormat.pixelFormat()][m_surfaceFormat];
     }
 
     [[nodiscard]] QSGMaterialShader *createShader(QSGRendererInterface::RenderMode) const override {
-        return new QSGVideoMaterialRhiShader(m_videoFormat, m_surfaceFormat, m_hdrInfo);
+        return new QSGVideoMaterialRhiShader(m_videoFormat, m_surfaceFormat, m_hdrInfo, m_rhi);
     }
 
     int compare(const QSGMaterial *other) const override {
@@ -95,11 +99,6 @@ public:
         setFlag(Blending, !qFuzzyCompare(m_opacity, float(1.0)));
     }
 
-    void setCurrentFrame(const QVideoFrame &frame) {
-        m_currentFrame = frame;
-        m_texturesDirty = true;
-    }
-
     void setSurfaceFormat(const QRhiSwapChain::Format surfaceFormat)
     {
         m_surfaceFormat = surfaceFormat;
@@ -117,32 +116,23 @@ public:
     float m_opacity = 1.0f;
     QRhiSwapChainHdrInfo m_hdrInfo;
 
-    bool m_texturesDirty = false;
-    QVideoFrame m_currentFrame;
-
-    enum { NVideoFrameSlots = 4 };
-    QVideoFrame m_videoFrameSlots[NVideoFrameSlots];
+    QVideoFrameTexturePoolPtr m_texturePool = std::make_shared<QVideoFrameTexturePool>();
     std::array<QSGVideoTexture, 3> m_textures;
-    std::unique_ptr<QVideoFrameTextures> m_videoFrameTextures;
+
+    QRhi *m_rhi;
 };
 
 void QSGVideoMaterial::updateTextures(QRhi *rhi, QRhiResourceUpdateBatch *resourceUpdates)
 {
-    if (!m_texturesDirty)
+    if (!m_texturePool->texturesDirty())
         return;
 
-    // keep the video frames alive until we know that they are not needed anymore
-    Q_ASSERT(NVideoFrameSlots >= rhi->resourceLimit(QRhi::FramesInFlight));
-    m_videoFrameSlots[rhi->currentFrameSlot()] = m_currentFrame;
-
-    // update and upload all textures
-    m_videoFrameTextures = QVideoTextureHelper::createTextures(m_currentFrame, rhi, resourceUpdates, std::move(m_videoFrameTextures));
-    if (!m_videoFrameTextures)
+    QVideoFrameTextures *textures = m_texturePool->updateTextures(*rhi, *resourceUpdates);
+    if (!textures)
         return;
 
     for (int plane = 0; plane < 3; ++plane)
-        m_textures[plane].setRhiTexture(m_videoFrameTextures->texture(plane));
-    m_texturesDirty = false;
+        m_textures[plane].setRhiTexture(textures->texture(plane));
 }
 
 
@@ -175,7 +165,8 @@ bool QSGVideoMaterialRhiShader::updateUniformData(RenderState &state, QSGMateria
     }
 
     QVideoTextureHelper::updateUniformData(state.uniformData(), m_videoFormat,
-        m->m_currentFrame, state.combinedMatrix(), state.opacity(), maxNits);
+                                           m->m_texturePool->currentFrame(), state.combinedMatrix(),
+                                           state.opacity(), maxNits);
 
     return true;
 }
@@ -192,18 +183,20 @@ void QSGVideoMaterialRhiShader::updateSampledImage(RenderState &state, int bindi
     *texture = &m->m_textures[binding - 1];
 }
 
-QSGVideoMaterial::QSGVideoMaterial(const QVideoFrameFormat &videoFormat)
-    : m_videoFormat(videoFormat)
+QSGVideoMaterial::QSGVideoMaterial(const QVideoFrameFormat &videoFormat, QRhi *rhi)
+    : m_videoFormat(videoFormat),
+    m_rhi(rhi)
 {
     setFlag(Blending, false);
 }
 
-QSGVideoNode::QSGVideoNode(QQuickVideoOutput *parent, const QVideoFrameFormat &videoFormat)
+QSGVideoNode::QSGVideoNode(QQuickVideoOutput *parent, const QVideoFrameFormat &videoFormat,
+                           QRhi *rhi)
     : m_parent(parent), m_videoFormat(videoFormat)
 {
     setFlag(QSGNode::OwnsMaterial);
     setFlag(QSGNode::OwnsGeometry);
-    m_material = new QSGVideoMaterial(videoFormat);
+    m_material = new QSGVideoMaterial(videoFormat, rhi);
     setMaterial(m_material);
 }
 
@@ -214,7 +207,7 @@ QSGVideoNode::~QSGVideoNode()
 
 void QSGVideoNode::setCurrentFrame(const QVideoFrame &frame)
 {
-    m_material->setCurrentFrame(frame);
+    texturePool()->setCurrentFrame(frame);
     markDirty(DirtyMaterial);
     updateSubtitle(frame);
 }
@@ -236,8 +229,9 @@ void QSGVideoNode::updateSubtitle(const QVideoFrame &frame)
     QSize subtitleFrameSize = m_rect.size().toSize();
     if (subtitleFrameSize.isEmpty())
         return;
-    if (m_orientation % 180)
-        subtitleFrameSize.transpose();
+
+    subtitleFrameSize = qRotatedFrameSize(subtitleFrameSize, m_videoOutputTransformation.rotation);
+
     if (!m_subtitleLayout.update(subtitleFrameSize, frame.subtitleText()))
         return;
 
@@ -246,11 +240,14 @@ void QSGVideoNode::updateSubtitle(const QVideoFrame &frame)
     if (frame.subtitleText().isEmpty())
         return;
 
-    m_subtitleTextNode = new QQuickTextNode(m_parent);
+    QQuickItemPrivate *parent_d = QQuickItemPrivate::get(m_parent);
+
+    m_subtitleTextNode = parent_d->sceneGraphContext()->createInternalTextNode(parent_d->sceneGraphRenderContext());
+    m_subtitleTextNode->setColor(Qt::white);
     QColor bgColor = Qt::black;
     bgColor.setAlpha(128);
     m_subtitleTextNode->addRectangleNode(m_subtitleLayout.bounds, bgColor);
-    m_subtitleTextNode->addTextLayout(m_subtitleLayout.layout.position(), &m_subtitleLayout.layout, Qt::white);
+    m_subtitleTextNode->addTextLayout(m_subtitleLayout.layout.position(), &m_subtitleLayout.layout);
     appendChildNode(m_subtitleTextNode);
     setSubtitleGeometry();
 }
@@ -261,52 +258,47 @@ void QSGVideoNode::setSubtitleGeometry()
         return;
 
     if (m_material)
-        updateSubtitle(m_material->m_currentFrame);
+        updateSubtitle(texturePool()->currentFrame());
 
-    float rotate = -1.f * m_orientation;
+    float rotate = -1.f * qToUnderlying(m_videoOutputTransformation.rotation);
     float yTranslate = 0;
     float xTranslate = 0;
-    if (m_orientation == 90) {
+    if (m_videoOutputTransformation.rotation == QtVideo::Rotation::Clockwise90) {
         yTranslate = m_rect.height();
-    } else if (m_orientation == 180) {
+    } else if (m_videoOutputTransformation.rotation == QtVideo::Rotation::Clockwise180) {
         yTranslate = m_rect.height();
         xTranslate = m_rect.width();
-    } else if (m_orientation == 270) {
+    } else if (m_videoOutputTransformation.rotation == QtVideo::Rotation::Clockwise270) {
         xTranslate = m_rect.width();
     }
 
     QMatrix4x4 transform;
     transform.translate(m_rect.x() + xTranslate, m_rect.y() + yTranslate);
     transform.rotate(rotate, 0, 0, 1);
+    // TODO: Investigate if we should we mirror subtitles
+    // if (m_videoOutputTransformation.mirrorredHorizontallyAfterRotation)
+    //    transform.scale(-1.f, 1.f);
 
     m_subtitleTextNode->setMatrix(transform);
     m_subtitleTextNode->markDirty(DirtyGeometry);
 }
 
-/* Update the vertices and texture coordinates.  Orientation must be in {0,90,180,270} */
-void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &textureRect, int orientation)
+/* Update the vertices and texture coordinates.*/
+void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &textureRect,
+                                           VideoTransformation videoOutputTransformation)
 {
-    bool frameChanged = false;
-    if (m_material) {
-        if (m_material->m_currentFrame.rotationAngle() != m_frameOrientation
-            || m_material->m_currentFrame.mirrored() != m_frameMirrored) {
-            frameChanged = true;
-        }
-    }
-    if (rect == m_rect && textureRect == m_textureRect && orientation == m_orientation
-        && !frameChanged)
+    const VideoTransformation currentFrameTransformation = qNormalizedFrameTransformation(
+            m_material ? texturePool()->currentFrame() : QVideoFrame{}, videoOutputTransformation);
+
+    if (rect == m_rect && textureRect == m_textureRect
+        && videoOutputTransformation == m_videoOutputTransformation
+        && currentFrameTransformation == m_frameTransformation)
         return;
 
     m_rect = rect;
     m_textureRect = textureRect;
-    m_orientation = orientation;
-    if (m_material) {
-        m_frameOrientation = m_material->m_currentFrame.rotationAngle();
-        m_frameMirrored = m_material->m_currentFrame.mirrored();
-    }
-    int videoRotation = orientation;
-    videoRotation += m_material ? m_material->m_currentFrame.rotationAngle() : 0;
-    videoRotation %= 360;
+    m_videoOutputTransformation = videoOutputTransformation;
+    m_frameTransformation = currentFrameTransformation;
 
     QSGGeometry *g = geometry();
 
@@ -315,6 +307,11 @@ void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &tex
 
     QSGGeometry::TexturedPoint2D *v = g->vertexDataAsTexturedPoint2D();
 
+    // Vertexes:
+    // 0   2
+    //
+    // 1   3
+
     // Set geometry first
     qSetGeom(v + 0, rect.topLeft());
     qSetGeom(v + 1, rect.bottomLeft());
@@ -322,7 +319,7 @@ void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &tex
     qSetGeom(v + 3, rect.bottomRight());
 
     // and then texture coordinates
-    switch (videoRotation) {
+    switch (currentFrameTransformation.rotation) {
     default:
         // tl, bl, tr, br
         qSetTex(v + 0, textureRect.topLeft());
@@ -331,7 +328,7 @@ void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &tex
         qSetTex(v + 3, textureRect.bottomRight());
         break;
 
-    case 90:
+    case QtVideo::Rotation::Clockwise90:
         // bl, br, tl, tr
         qSetTex(v + 0, textureRect.bottomLeft());
         qSetTex(v + 1, textureRect.bottomRight());
@@ -339,7 +336,7 @@ void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &tex
         qSetTex(v + 3, textureRect.topRight());
         break;
 
-    case 180:
+    case QtVideo::Rotation::Clockwise180:
         // br, tr, bl, tl
         qSetTex(v + 0, textureRect.bottomRight());
         qSetTex(v + 1, textureRect.topRight());
@@ -347,7 +344,7 @@ void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &tex
         qSetTex(v + 3, textureRect.topLeft());
         break;
 
-    case 270:
+    case QtVideo::Rotation::Clockwise270:
         // tr, tl, br, bl
         qSetTex(v + 0, textureRect.topRight());
         qSetTex(v + 1, textureRect.topLeft());
@@ -356,7 +353,7 @@ void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &tex
         break;
     }
 
-    if (m_material && m_material->m_currentFrame.mirrored()) {
+    if (m_frameTransformation.mirrorredHorizontallyAfterRotation) {
         qSwapTex(v + 0, v + 2);
         qSwapTex(v + 1, v + 3);
     }
@@ -369,4 +366,9 @@ void QSGVideoNode::setTexturedRectGeometry(const QRectF &rect, const QRectF &tex
     setSubtitleGeometry();
 }
 
+const QVideoFrameTexturePoolPtr &QSGVideoNode::texturePool() const
+{
+    return m_material->m_texturePool;
+}
+
 QT_END_NAMESPACE
