@@ -8,11 +8,12 @@
 #include "qandroidaudiooutput_p.h"
 #include "qaudiooutput.h"
 
+#include <private/qplatformvideosink_p.h>
 #include <qloggingcategory.h>
 
 QT_BEGIN_NAMESPACE
 
-Q_LOGGING_CATEGORY(lcMediaPlayer, "qt.multimedia.mediaplayer.android")
+static Q_LOGGING_CATEGORY(lcMediaPlayer, "qt.multimedia.mediaplayer.android")
 
 class StateChangeNotifier
 {
@@ -68,6 +69,9 @@ QAndroidMediaPlayer::QAndroidMediaPlayer(QMediaPlayer *parent)
 
 QAndroidMediaPlayer::~QAndroidMediaPlayer()
 {
+    if (m_videoSink)
+        disconnect(m_videoSink->platformVideoSink(), nullptr, this, nullptr);
+
     mMediaPlayer->disconnect();
     mMediaPlayer->release();
     delete mMediaPlayer;
@@ -272,7 +276,8 @@ void QAndroidMediaPlayer::setMedia(const QUrl &mediaContent,
         if (mVideoSize.isValid() && mVideoOutput)
             mVideoOutput->setVideoSize(mVideoSize);
 
-        if ((mMediaPlayer->display() == 0) && mVideoOutput)
+        if (mVideoOutput &&
+                        (mMediaPlayer->display() == 0 || mVideoOutput->shouldTextureBeUpdated()))
             mMediaPlayer->setDisplay(mVideoOutput->surfaceTexture());
         mMediaPlayer->setDataSource(QNetworkRequest(mediaContent));
         mMediaPlayer->prepareAsync();
@@ -291,6 +296,9 @@ void QAndroidMediaPlayer::setVideoSink(QVideoSink *sink)
     if (m_videoSink == sink)
         return;
 
+    if (m_videoSink)
+        disconnect(m_videoSink->platformVideoSink(), nullptr, this, nullptr);
+
     m_videoSink = sink;
 
     if (!m_videoSink) {
@@ -311,6 +319,9 @@ void QAndroidMediaPlayer::setVideoSink(QVideoSink *sink)
 
     if (mVideoOutput->isReady())
         mMediaPlayer->setDisplay(mVideoOutput->surfaceTexture());
+
+    connect(m_videoSink->platformVideoSink(), &QPlatformVideoSink::rhiChanged, this, [&]()
+            { mMediaPlayer->setDisplay(mVideoOutput->surfaceTexture()); });
 }
 
 void QAndroidMediaPlayer::setAudioOutput(QPlatformAudioOutput *output)
@@ -502,7 +513,9 @@ void QAndroidMediaPlayer::onError(qint32 what, qint32 extra)
         setMediaStatus(QMediaPlayer::InvalidMedia);
         break;
     case AndroidMediaPlayer::MEDIA_ERROR_BAD_THINGS_ARE_GOING_TO_HAPPEN:
-        errorString += QLatin1String(" (Unknown error/Insufficient resources)");
+        errorString += mMediaContent.scheme() == QLatin1String("rtsp")
+            ? QLatin1String(" (Unknown error/Insufficient resources or RTSP may not be supported)")
+            : QLatin1String(" (Unknown error/Insufficient resources)");
         error = QMediaPlayer::ResourceError;
         break;
     }
@@ -635,7 +648,6 @@ void QAndroidMediaPlayer::onStateChanged(qint32 state)
         mMediaPlayer->setDisplay(0);
         if (mVideoOutput) {
             mVideoOutput->stop();
-            mVideoOutput->reset();
         }
     }
 }
@@ -983,3 +995,5 @@ void QAndroidMediaPlayer::updateTrackInfo()
 }
 
 QT_END_NAMESPACE
+
+#include "moc_qandroidmediaplayer_p.cpp"
