@@ -8,6 +8,8 @@
 #include "qandroidmultimediautils_p.h"
 #include "qandroidglobal_p.h"
 
+#include <private/qvideoframe_p.h>
+
 #include <qhash.h>
 #include <qstringlist.h>
 #include <qdebug.h>
@@ -21,7 +23,7 @@
 
 QT_BEGIN_NAMESPACE
 
-Q_LOGGING_CATEGORY(lcAndroidCamera, "qt.multimedia.android.camera")
+static Q_LOGGING_CATEGORY(lcAndroidCamera, "qt.multimedia.android.camera")
 
 static const char QtCameraListenerClassName[] = "org/qtproject/qt/android/multimedia/QtCameraListener";
 
@@ -125,10 +127,9 @@ static void notifyPictureCaptured(JNIEnv *env, jobject, int id, jbyteArray data)
         bytesPerLine = -1;
     }
 
-    QVideoFrame frame(new QMemoryVideoBuffer(bytes, bytesPerLine),
-                      QVideoFrameFormat(pictureSize, qt_pixelFormatFromAndroidImageFormat(format)));
+    auto pictureFormat = qt_pixelFormatFromAndroidImageFormat(format);
 
-    emit camera->pictureCaptured(frame);
+    emit camera->pictureCaptured(bytes, pictureFormat, pictureSize, bytesPerLine);
 }
 
 static void notifyNewPreviewFrame(JNIEnv *env, jobject, int id, jbyteArray data,
@@ -146,9 +147,12 @@ static void notifyNewPreviewFrame(JNIEnv *env, jobject, int id, jbyteArray data,
     QByteArray bytes(arrayLength, Qt::Uninitialized);
     env->GetByteArrayRegion(data, 0, arrayLength, (jbyte*)bytes.data());
 
-    QVideoFrame frame(new QMemoryVideoBuffer(bytes, bpl),
-                      QVideoFrameFormat(QSize(width, height),
-                                          qt_pixelFormatFromAndroidImageFormat(AndroidCamera::ImageFormat(format))));
+    QVideoFrameFormat frameFormat(
+            QSize(width, height),
+            qt_pixelFormatFromAndroidImageFormat(AndroidCamera::ImageFormat(format)));
+
+    QVideoFrame frame = QVideoFramePrivate::createFrame(
+            std::make_unique<QMemoryVideoBuffer>(std::move(bytes), bpl), std::move(frameFormat));
 
     Q_EMIT (*it)->newPreviewFrame(frame);
 }
@@ -316,7 +320,7 @@ AndroidCamera::~AndroidCamera()
 
 AndroidCamera *AndroidCamera::open(int cameraId)
 {
-    if (!qt_androidRequestCameraPermission())
+    if (!qt_androidCheckCameraPermission())
         return nullptr;
 
     AndroidCameraPrivate *d = new AndroidCameraPrivate();
@@ -765,7 +769,7 @@ QJniObject AndroidCamera::getCameraObject()
 
 int AndroidCamera::getNumberOfCameras()
 {
-    if (!qt_androidRequestCameraPermission())
+    if (!qt_androidCheckCameraPermission())
         return 0;
 
     return QJniObject::callStaticMethod<jint>("android/hardware/Camera",
@@ -801,6 +805,12 @@ void AndroidCamera::getCameraInfo(int id, QCameraDevicePrivate *info)
     default:
         break;
     }
+    // Add a number to allow correct access to cameras on systems with two
+    // (and more) front/back cameras
+    if (id > 1) {
+        info->id.append(QByteArray::number(id));
+        info->description.append(QString(" %1").arg(id));
+    }
 }
 
 QVideoFrameFormat::PixelFormat AndroidCamera::QtPixelFormatFromAndroidImageFormat(AndroidCamera::ImageFormat format)
@@ -1199,6 +1209,7 @@ bool AndroidCameraPrivate::setPreviewDisplay(void *surfaceHolder)
 void AndroidCameraPrivate::setDisplayOrientation(int degrees)
 {
     m_camera.callMethod<void>("setDisplayOrientation", "(I)V", degrees);
+    m_cameraListener.callMethod<void>("setPhotoRotation", "(I)V", degrees);
 }
 
 bool AndroidCameraPrivate::isZoomSupported()
@@ -1724,9 +1735,12 @@ void AndroidCameraPrivate::fetchLastPreviewFrame()
     const int format = m_cameraListener.callMethod<jint>("previewFormat");
     const int bpl = m_cameraListener.callMethod<jint>("previewBytesPerLine");
 
-    QVideoFrame frame(new QMemoryVideoBuffer(bytes, bpl),
-                      QVideoFrameFormat(QSize(width, height),
-                                          qt_pixelFormatFromAndroidImageFormat(AndroidCamera::ImageFormat(format))));
+    QVideoFrameFormat frameFormat(
+            QSize(width, height),
+            qt_pixelFormatFromAndroidImageFormat(AndroidCamera::ImageFormat(format)));
+
+    QVideoFrame frame = QVideoFramePrivate::createFrame(
+            std::make_unique<QMemoryVideoBuffer>(std::move(bytes), bpl), std::move(frameFormat));
 
     emit lastPreviewFrameFetched(frame);
 }
@@ -1780,3 +1794,4 @@ bool AndroidCamera::registerNativeMethods()
 QT_END_NAMESPACE
 
 #include "androidcamera.moc"
+#include "moc_androidcamera_p.cpp"
