@@ -16,10 +16,12 @@
 #include "private/qmediarecorder_p.h"
 #include "qdarwinformatsinfo_p.h"
 #include "private/qplatformaudiooutput_p.h"
+#include <private/qplatformaudioinput_p.h>
 
 #include <QtCore/qmath.h>
 #include <QtCore/qdebug.h>
-#include <QtCore/qmimetype.h>
+
+#include <private/qcoreaudioutils_p.h>
 
 QT_USE_NAMESPACE
 
@@ -99,7 +101,7 @@ void AVFMediaEncoder::updateDuration(qint64 duration)
     durationChanged(m_duration);
 }
 
-static NSDictionary *avfAudioSettings(const QMediaEncoderSettings &encoderSettings)
+static NSDictionary *avfAudioSettings(const QMediaEncoderSettings &encoderSettings, const QAudioFormat &format)
 {
     NSMutableDictionary *settings = [NSMutableDictionary dictionary];
 
@@ -183,22 +185,38 @@ static NSDictionary *avfAudioSettings(const QMediaEncoderSettings &encoderSettin
                 }
             }
         }
+
+        // if channel count is provided and it's bigger than 2
+        // provide a supported channel layout
+        if (isChannelCountSupported && channelCount > 2) {
+            AudioChannelLayout channelLayout;
+            memset(&channelLayout, 0, sizeof(AudioChannelLayout));
+            auto channelLayoutTags = qt_supported_channel_layout_tags_for_format(codecId, channelCount);
+            if (channelLayoutTags.size()) {
+                channelLayout.mChannelLayoutTag = channelLayoutTags.first();
+                [settings setObject:[NSData dataWithBytes: &channelLayout length: sizeof(channelLayout)] forKey:AVChannelLayoutKey];
+            } else {
+                isChannelCountSupported = false;
+            }
+        }
+
+        if (isChannelCountSupported)
+            [settings setObject:[NSNumber numberWithInt:channelCount] forKey:AVNumberOfChannelsKey];
     }
 
-    if (isChannelCountSupported && channelCount > 2) {
-        AudioChannelLayout channelLayout;
-        memset(&channelLayout, 0, sizeof(AudioChannelLayout));
-        auto channelLayoutTags = qt_supported_channel_layout_tags_for_format(codecId, channelCount);
-        if (channelLayoutTags.size()) {
-            channelLayout.mChannelLayoutTag = channelLayoutTags.first();
-            [settings setObject:[NSData dataWithBytes: &channelLayout length: sizeof(channelLayout)] forKey:AVChannelLayoutKey];
+    if (!isChannelCountSupported) {
+        // fallback to providing channel layout if channel count is not specified or supported
+        UInt32 size = 0;
+        if (format.isValid()) {
+            auto layout = QCoreAudioUtils::toAudioChannelLayout(format, &size);
+            UInt32 layoutSize = offsetof(AudioChannelLayout, mChannelDescriptions)
+                    + layout->mNumberChannelDescriptions * sizeof(AudioChannelDescription);
+            [settings setObject:[NSData dataWithBytes:layout.get() length:layoutSize] forKey:AVChannelLayoutKey];
         } else {
-            isChannelCountSupported = false;
+            // finally default to setting channel count to 1
+            [settings setObject:[NSNumber numberWithInt:1] forKey:AVNumberOfChannelsKey];
         }
     }
-    if (!isChannelCountSupported)
-        channelCount = 2;
-    [settings setObject:[NSNumber numberWithInt:channelCount] forKey:AVNumberOfChannelsKey];
 
     if (codecId == kAudioFormatAppleLossless)
         [settings setObject:[NSNumber numberWithInt:24] forKey:AVEncoderBitDepthHintKey];
@@ -385,7 +403,9 @@ void AVFMediaEncoder::applySettings(QMediaEncoderSettings &settings)
     AVFCameraSession *session = m_service->session();
 
     // audio settings
-    m_audioSettings = avfAudioSettings(settings);
+    const auto audioInput = m_service->audioInput();
+    const QAudioFormat audioFormat = audioInput ? audioInput->device.preferredFormat() : QAudioFormat();
+    m_audioSettings = avfAudioSettings(settings, audioFormat);
     if (m_audioSettings)
         [m_audioSettings retain];
 
@@ -460,7 +480,7 @@ void AVFMediaEncoder::record(QMediaEncoderSettings &settings)
 
     if (!cameraControl && !audioInput) {
         qWarning() << Q_FUNC_INFO << "Cannot record without any inputs";
-        Q_EMIT error(QMediaRecorder::ResourceError, tr("No inputs specified"));
+        updateError(QMediaRecorder::ResourceError, tr("No inputs specified"));
         return;
     }
 
@@ -472,28 +492,28 @@ void AVFMediaEncoder::record(QMediaEncoderSettings &settings)
     if (!audioOnly) {
         if (!cameraControl || !cameraControl->isActive()) {
             qCDebug(qLcCamera) << Q_FUNC_INFO << "can not start record while camera is not active";
-            Q_EMIT error(QMediaRecorder::ResourceError,
-                         QMediaRecorderPrivate::msgFailedStartRecording());
+            updateError(QMediaRecorder::ResourceError,
+                        QMediaRecorderPrivate::msgFailedStartRecording());
             return;
         }
     }
 
     const QString path(outputLocation().scheme() == QLatin1String("file") ?
                            outputLocation().path() : outputLocation().toString());
-    const QUrl fileURL(QUrl::fromLocalFile(QMediaStorageLocation::generateFileName(path,
-                    audioOnly ? QStandardPaths::MusicLocation : QStandardPaths::MoviesLocation,
-                    settings.mimeType().preferredSuffix())));
+    const QUrl fileURL(QUrl::fromLocalFile(QMediaStorageLocation::generateFileName(
+            path, audioOnly ? QStandardPaths::MusicLocation : QStandardPaths::MoviesLocation,
+            settings.preferredSuffix())));
 
     NSURL *nsFileURL = fileURL.toNSURL();
     if (!nsFileURL) {
         qWarning() << Q_FUNC_INFO << "invalid output URL:" << fileURL;
-        Q_EMIT error(QMediaRecorder::ResourceError, tr("Invalid output file URL"));
+        updateError(QMediaRecorder::ResourceError, tr("Invalid output file URL"));
         return;
     }
     if (!qt_is_writable_file_URL(nsFileURL)) {
         qWarning() << Q_FUNC_INFO << "invalid output URL:" << fileURL
                     << "(the location is not writable)";
-        Q_EMIT error(QMediaRecorder::ResourceError, tr("Non-writeable file location"));
+        updateError(QMediaRecorder::ResourceError, tr("Non-writeable file location"));
         return;
     }
     if (qt_file_exists(nsFileURL)) {
@@ -501,7 +521,7 @@ void AVFMediaEncoder::record(QMediaEncoderSettings &settings)
         // Objective-C exception, which is not good at all.
         qWarning() << Q_FUNC_INFO << "invalid output URL:" << fileURL
                     << "(file already exists)";
-        Q_EMIT error(QMediaRecorder::ResourceError, tr("File already exists"));
+        updateError(QMediaRecorder::ResourceError, tr("File already exists"));
         return;
     }
 
@@ -536,8 +556,7 @@ void AVFMediaEncoder::record(QMediaEncoderSettings &settings)
         [m_writer start];
     } else {
         [session startRunning];
-        Q_EMIT error(QMediaRecorder::FormatError,
-                     QMediaRecorderPrivate::msgFailedStartRecording());
+        updateError(QMediaRecorder::FormatError, QMediaRecorderPrivate::msgFailedStartRecording());
     }
 }
 
@@ -602,7 +621,8 @@ void AVFMediaEncoder::assetWriterFinished()
         if (session->audioPreviewDelegate()) {
             [session->audioPreviewDelegate() resetAudioPreviewDelegate];
         }
-        [session->captureSession() startRunning];
+        if (session->videoOutput() || session->audioPreviewDelegate())
+            [session->captureSession() startRunning];
     }
 
     m_state = QMediaRecorder::StoppedState;
@@ -612,7 +632,7 @@ void AVFMediaEncoder::assetWriterFinished()
 
 void AVFMediaEncoder::assetWriterError(QString err)
 {
-    Q_EMIT error(QMediaRecorder::FormatError, err);
+    updateError(QMediaRecorder::FormatError, err);
     if (m_state != QMediaRecorder::StoppedState)
         stopWriter();
 }
