@@ -25,11 +25,17 @@ class MediaCodecTextureConverter : public TextureConverterBackend
 {
 public:
     MediaCodecTextureConverter(QRhi *rhi) : TextureConverterBackend(rhi){};
-    TextureSet *getTextures(AVFrame *frame) override;
+
+    QVideoFrameTexturesUPtr createTextures(AVFrame *frame,
+                                           QVideoFrameTexturesUPtr &oldTextures) override;
+
+    QVideoFrameTexturesHandlesUPtr
+    createTextureHandles(AVFrame *frame, QVideoFrameTexturesHandlesUPtr oldHandles) override;
 
     static void setupDecoderSurface(AVCodecContext *s);
 private:
     std::unique_ptr<QRhiTexture> externalTexture;
+    quint64 m_currentSurfaceIndex = 0;
 };
 }
 #endif // QFFMPEGHWACCEL_MEDIACODEC_P_H
