@@ -1,59 +1,60 @@
 // Copyright (C) 2016 Jolla Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
-#include <qvideoframe.h>
-#include <qvideosink.h>
-#include <QDebug>
-#include <QMap>
-#include <QThread>
-#include <QEvent>
-#include <QCoreApplication>
-
-#include <private/qfactoryloader_p.h>
-#include "qgstvideobuffer_p.h"
-#include "qgstreamervideosink_p.h"
-
 #include "qgstvideorenderersink_p.h"
 
+#include <QtMultimedia/qvideoframe.h>
+#include <QtMultimedia/qvideosink.h>
+#include <QtMultimedia/private/qvideoframe_p.h>
+#include <QtGui/rhi/qrhi.h>
+#include <QtCore/qcoreapplication.h>
+#include <QtCore/qdebug.h>
+#include <QtCore/qloggingcategory.h>
+#include <QtCore/private/qfactoryloader_p.h>
+#include <QtCore/private/quniquehandle_p.h>
+
+#include <common/qgst_debug_p.h>
+#include <common/qgstreamermetadata_p.h>
+#include <common/qgstreamervideosink_p.h>
+#include <common/qgstutils_p.h>
+#include <common/qgstvideobuffer_p.h>
+
 #include <gst/video/video.h>
 #include <gst/video/gstvideometa.h>
-#include <qloggingcategory.h>
-#include <qdebug.h>
 
-#include "qgstutils_p.h"
 
-#include <rhi/qrhi.h>
 #if QT_CONFIG(gstreamer_gl)
 #include <gst/gl/gl.h>
 #endif // #if QT_CONFIG(gstreamer_gl)
 
 // DMA support
-#if QT_CONFIG(linux_dmabuf)
-#include <gst/allocators/gstdmabuf.h>
+#if QT_CONFIG(gstreamer_gl_egl) && QT_CONFIG(linux_dmabuf)
+#  include <gst/allocators/gstdmabuf.h>
 #endif
 
-//#define DEBUG_VIDEO_SURFACE_SINK
+// NOLINTBEGIN(readability-convert-member-functions-to-static)
 
 static Q_LOGGING_CATEGORY(qLcGstVideoRenderer, "qt.multimedia.gstvideorenderer")
 
 QT_BEGIN_NAMESPACE
 
 QGstVideoRenderer::QGstVideoRenderer(QGstreamerVideoSink *sink)
-    : m_sink(sink)
+    : m_sink(sink), m_surfaceCaps(createSurfaceCaps(sink))
 {
-    createSurfaceCaps();
+    QObject::connect(
+            sink, &QGstreamerVideoSink::aboutToBeDestroyed, this,
+            [this] {
+                QMutexLocker locker(&m_sinkMutex);
+                m_sink = nullptr;
+            },
+            Qt::DirectConnection);
 }
 
-QGstVideoRenderer::~QGstVideoRenderer()
-{
-}
+QGstVideoRenderer::~QGstVideoRenderer() = default;
 
-void QGstVideoRenderer::createSurfaceCaps()
+QGstCaps QGstVideoRenderer::createSurfaceCaps([[maybe_unused]] QGstreamerVideoSink *sink)
 {
-    QRhi *rhi = m_sink->rhi();
-    Q_UNUSED(rhi);
-
-    auto caps = QGstCaps::create();
+    QGstCaps caps = QGstCaps::create();
 
     // All the formats that both we and gstreamer support
     auto formats = QList<QVideoFrameFormat::PixelFormat>()
@@ -78,10 +79,11 @@ void QGstVideoRenderer::createSurfaceCaps()
                    << QVideoFrameFormat::Format_Y16
         ;
 #if QT_CONFIG(gstreamer_gl)
+    QRhi *rhi = sink->rhi();
     if (rhi && rhi->backend() == QRhi::OpenGLES2) {
         caps.addPixelFormats(formats, GST_CAPS_FEATURE_MEMORY_GL_MEMORY);
-#if QT_CONFIG(linux_dmabuf)
-        if (m_sink->eglDisplay() && m_sink->eglImageTargetTexture2D()) {
+#  if QT_CONFIG(gstreamer_gl_egl) && QT_CONFIG(linux_dmabuf)
+        if (sink->eglDisplay() && sink->eglImageTargetTexture2D()) {
             // We currently do not handle planar DMA buffers, as it's somewhat unclear how to
             // convert the planar EGLImage into something we can use from OpenGL
             auto singlePlaneFormats = QList<QVideoFrameFormat::PixelFormat>()
@@ -101,109 +103,134 @@ void QGstVideoRenderer::createSurfaceCaps()
                 ;
             caps.addPixelFormats(singlePlaneFormats, GST_CAPS_FEATURE_MEMORY_DMABUF);
         }
-#endif
+#  endif
     }
 #endif
     caps.addPixelFormats(formats);
-
-    m_surfaceCaps = caps;
+    return caps;
 }
 
-QGstCaps QGstVideoRenderer::caps()
+void QGstVideoRenderer::customEvent(QEvent *event)
 {
-    QMutexLocker locker(&m_mutex);
+QT_WARNING_PUSH
+QT_WARNING_DISABLE_GCC("-Wswitch") // case value not in enumerated type ‘QEvent::Type’
 
+    switch (event->type()) {
+    case renderFramesEvent: {
+        // LATER: we currently show every frame. however it may be reasonable to drop frames
+        // here if the queue contains more than one frame
+        while (std::optional<RenderBufferState> nextState = m_bufferQueue.dequeue())
+            handleNewBuffer(std::move(*nextState));
+        return;
+    }
+    case stopEvent: {
+        m_currentPipelineFrame = {};
+        updateCurrentVideoFrame(m_currentVideoFrame);
+        return;
+    }
+
+    default:
+        return;
+    }
+QT_WARNING_POP
+}
+
+
+void QGstVideoRenderer::handleNewBuffer(RenderBufferState state)
+{
+    auto videoBuffer = std::make_unique<QGstVideoBuffer>(state.buffer, m_videoInfo, m_sink,
+                                                         state.format, state.memoryFormat);
+    QVideoFrame frame = QVideoFramePrivate::createFrame(std::move(videoBuffer), state.format);
+    QGstUtils::setFrameTimeStampsFromBuffer(&frame, state.buffer.get());
+    m_currentPipelineFrame = std::move(frame);
+
+    if (!m_isActive) {
+        qCDebug(qLcGstVideoRenderer) << "    showing empty video frame";
+        updateCurrentVideoFrame({});
+        return;
+    }
+
+    updateCurrentVideoFrame(m_currentPipelineFrame);
+}
+
+const QGstCaps &QGstVideoRenderer::caps()
+{
     return m_surfaceCaps;
 }
 
 bool QGstVideoRenderer::start(const QGstCaps& caps)
 {
-    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::start" << caps.toString();
-    QMutexLocker locker(&m_mutex);
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::start" << caps;
 
-    m_frameMirrored = false;
-    m_frameRotationAngle = QVideo::Rotation0;
-
-    if (m_active) {
-        m_flush = true;
-        m_stop = true;
+    auto optionalFormatAndVideoInfo = caps.formatAndVideoInfo();
+    if (optionalFormatAndVideoInfo) {
+        std::tie(m_format, m_videoInfo) = std::move(*optionalFormatAndVideoInfo);
+    } else {
+        m_format = {};
+        m_videoInfo = {};
     }
+    m_memoryFormat = caps.memoryFormat();
 
-    m_startCaps = caps;
+    // NOTE: m_format will not be fully populated until GST_EVENT_TAG is processed
 
-    /*
-    Waiting for start() to be invoked in the main thread may block
-    if gstreamer blocks the main thread until this call is finished.
-    This situation is rare and usually caused by setState(Null)
-    while pipeline is being prerolled.
-
-    The proper solution to this involves controlling gstreamer pipeline from
-    other thread than video surface.
-
-    Currently start() fails if wait() timed out.
-    */
-    if (!waitForAsyncEvent(&locker, &m_setupCondition, 1000) && !m_startCaps.isNull()) {
-        qWarning() << "Failed to start video surface due to main thread blocked.";
-        m_startCaps = {};
-    }
-
-    return m_active;
+    return true;
 }
 
 void QGstVideoRenderer::stop()
 {
-    QMutexLocker locker(&m_mutex);
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::stop";
 
-    if (!m_active)
-        return;
-
-    m_flush = true;
-    m_stop = true;
-
-    m_startCaps = {};
-
-    waitForAsyncEvent(&locker, &m_setupCondition, 500);
+    m_bufferQueue.clear();
+    QCoreApplication::postEvent(this, new QEvent(stopEvent));
 }
 
 void QGstVideoRenderer::unlock()
 {
-    QMutexLocker locker(&m_mutex);
-
-    m_setupCondition.wakeAll();
-    m_renderCondition.wakeAll();
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::unlock";
 }
 
-bool QGstVideoRenderer::proposeAllocation(GstQuery *query)
+bool QGstVideoRenderer::proposeAllocation(GstQuery *)
 {
-    Q_UNUSED(query);
-    QMutexLocker locker(&m_mutex);
-    return m_active;
-}
-
-void QGstVideoRenderer::flush()
-{
-    QMutexLocker locker(&m_mutex);
-
-    m_flush = true;
-    m_renderBuffer = nullptr;
-    m_renderCondition.wakeAll();
-
-    notify();
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::proposeAllocation";
+    return true;
 }
 
 GstFlowReturn QGstVideoRenderer::render(GstBuffer *buffer)
 {
-    QMutexLocker locker(&m_mutex);
     qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::render";
 
-    m_renderReturn = GST_FLOW_OK;
-    m_renderBuffer = buffer;
+    if (m_flushing) {
+        qCDebug(qLcGstVideoRenderer)
+                << "    buffer received while flushing the sink ... discarding buffer";
+        return GST_FLOW_FLUSHING;
+    }
 
-    waitForAsyncEvent(&locker, &m_renderCondition, 300);
+    GstVideoCropMeta *meta = gst_buffer_get_video_crop_meta(buffer);
+    if (meta) {
+        QRect vp(meta->x, meta->y, meta->width, meta->height);
+        if (m_format.viewport() != vp) {
+            qCDebug(qLcGstVideoRenderer)
+                    << Q_FUNC_INFO << " Update viewport on Metadata: [" << meta->height << "x"
+                    << meta->width << " | " << meta->x << "x" << meta->y << "]";
+            // Update viewport if data is not the same
+            m_format.setViewport(vp);
+        }
+    }
 
-    m_renderBuffer = nullptr;
+    RenderBufferState state{
+        QGstBufferHandle{ buffer, QGstBufferHandle::NeedsRef },
+        m_format,
+        m_memoryFormat,
+    };
 
-    return m_renderReturn;
+    qCDebug(qLcGstVideoRenderer) << "    sending video frame";
+
+    qsizetype sizeOfQueue = m_bufferQueue.enqueue(std::move(state));
+    if (sizeOfQueue == 1)
+        // we only need to wake up, if we don't have a pending frame
+        QCoreApplication::postEvent(this, new QEvent(renderFramesEvent));
+
+    return GST_FLOW_OK;
 }
 
 bool QGstVideoRenderer::query(GstQuery *query)
@@ -216,6 +243,10 @@ bool QGstVideoRenderer::query(GstQuery *query)
         if (strcmp(type, "gst.gl.local_context") != 0)
             return false;
 
+        QMutexLocker locker(&m_sinkMutex);
+        if (!m_sink)
+            return false;
+
         auto *gstGlContext = m_sink->gstGlLocalContext();
         if (!gstGlContext)
             return false;
@@ -232,174 +263,77 @@ bool QGstVideoRenderer::query(GstQuery *query)
 
 void QGstVideoRenderer::gstEvent(GstEvent *event)
 {
-    if (GST_EVENT_TYPE(event) != GST_EVENT_TAG)
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::gstEvent:" << event;
+
+    switch (GST_EVENT_TYPE(event)) {
+    case GST_EVENT_TAG:
+        return gstEventHandleTag(event);
+    case GST_EVENT_EOS:
+        return gstEventHandleEOS(event);
+    case GST_EVENT_FLUSH_START:
+        return gstEventHandleFlushStart(event);
+    case GST_EVENT_FLUSH_STOP:
+        return gstEventHandleFlushStop(event);
+
+    default:
+        return;
+    }
+}
+
+void QGstVideoRenderer::setActive(bool isActive)
+{
+    if (isActive == m_isActive)
         return;
 
+    m_isActive = isActive;
+    if (isActive)
+        updateCurrentVideoFrame(m_currentPipelineFrame);
+    else
+        updateCurrentVideoFrame({});
+}
+
+void QGstVideoRenderer::updateCurrentVideoFrame(QVideoFrame frame)
+{
+    m_currentVideoFrame = std::move(frame);
+    if (m_sink)
+        m_sink->setVideoFrame(m_currentVideoFrame);
+}
+
+void QGstVideoRenderer::gstEventHandleTag(GstEvent *event)
+{
     GstTagList *taglist = nullptr;
     gst_event_parse_tag(event, &taglist);
     if (!taglist)
         return;
 
-    gchar *value = nullptr;
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::gstEventHandleTag:" << taglist;
+
+    QGString value;
     if (!gst_tag_list_get_string(taglist, GST_TAG_IMAGE_ORIENTATION, &value))
         return;
 
-    constexpr const char rotate[] = "rotate-";
-    constexpr const char flipRotate[] = "flip-rotate-";
-    constexpr size_t rotateLen = sizeof(rotate) - 1;
-    constexpr size_t flipRotateLen = sizeof(flipRotate) - 1;
+    RotationResult parsed = parseRotationTag(value.get());
 
-    bool mirrored = false;
-    int rotationAngle = 0;
-
-    if (!strncmp(rotate, value, rotateLen)) {
-        rotationAngle = atoi(value + rotateLen);
-    } else if (!strncmp(flipRotate, value, flipRotateLen)) {
-        // To flip by horizontal axis is the same as to mirror by vertical axis
-        // and rotate by 180 degrees.
-        mirrored = true;
-        rotationAngle = (180 + atoi(value + flipRotateLen)) % 360;
-    }
-
-    QMutexLocker locker(&m_mutex);
-    m_frameMirrored = mirrored;
-    switch (rotationAngle) {
-    case 0: m_frameRotationAngle = QVideo::Rotation0; break;
-    case 90: m_frameRotationAngle = QVideo::Rotation90; break;
-    case 180: m_frameRotationAngle = QVideo::Rotation180; break;
-    case 270: m_frameRotationAngle = QVideo::Rotation270; break;
-    default: m_frameRotationAngle = QVideo::Rotation0;
-    }
+    m_format.setMirrored(parsed.flip);
+    m_format.setRotation(parsed.rotation);
 }
 
-bool QGstVideoRenderer::event(QEvent *event)
+void QGstVideoRenderer::gstEventHandleEOS(GstEvent *)
 {
-    if (event->type() == QEvent::UpdateRequest) {
-        QMutexLocker locker(&m_mutex);
-
-        if (m_notified) {
-            while (handleEvent(&locker)) {}
-            m_notified = false;
-        }
-        return true;
-    }
-
-    return QObject::event(event);
+    stop();
 }
 
-bool QGstVideoRenderer::handleEvent(QMutexLocker<QMutex> *locker)
+void QGstVideoRenderer::gstEventHandleFlushStart(GstEvent *)
 {
-    if (m_flush) {
-        m_flush = false;
-        if (m_active) {
-            locker->unlock();
-
-            if (m_sink && !m_flushed)
-                m_sink->setVideoFrame(QVideoFrame());
-            m_flushed = true;
-            locker->relock();
-        }
-    } else if (m_stop) {
-        m_stop = false;
-
-        if (m_active) {
-            m_active = false;
-            m_flushed = true;
-        }
-    } else if (!m_startCaps.isNull()) {
-        Q_ASSERT(!m_active);
-
-        auto startCaps = m_startCaps;
-        m_startCaps = {};
-
-        if (m_sink) {
-            locker->unlock();
-
-            m_flushed = true;
-            m_format = startCaps.formatForCaps(&m_videoInfo);
-            memoryFormat = startCaps.memoryFormat();
-
-            locker->relock();
-            m_active = m_format.isValid();
-        } else if (m_active) {
-            m_active = false;
-            m_flushed = true;
-        }
-
-    } else if (m_renderBuffer) {
-        GstBuffer *buffer = m_renderBuffer;
-        m_renderBuffer = nullptr;
-        m_renderReturn = GST_FLOW_ERROR;
-
-        qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::handleEvent(renderBuffer)" << m_active << m_sink;
-        if (m_active && m_sink) {
-            gst_buffer_ref(buffer);
-
-            locker->unlock();
-
-            m_flushed = false;
-
-            auto meta = gst_buffer_get_video_crop_meta (buffer);
-            if (meta) {
-                QRect vp(meta->x, meta->y, meta->width, meta->height);
-                if (m_format.viewport() != vp) {
-                    qCDebug(qLcGstVideoRenderer) << Q_FUNC_INFO << " Update viewport on Metadata: [" << meta->height << "x" << meta->width << " | " << meta->x << "x" << meta->y << "]";
-                    // Update viewport if data is not the same
-                    m_format.setViewport(vp);
-                }
-            }
-
-            if (m_sink->inStoppedState()) {
-                qCDebug(qLcGstVideoRenderer) << "    sending empty video frame";
-                m_sink->setVideoFrame(QVideoFrame());
-            } else {
-                QGstVideoBuffer *videoBuffer = new QGstVideoBuffer(buffer, m_videoInfo, m_sink, m_format, memoryFormat);
-                QVideoFrame frame(videoBuffer, m_format);
-                QGstUtils::setFrameTimeStamps(&frame, buffer);
-                frame.setMirrored(m_frameMirrored);
-                frame.setRotationAngle(m_frameRotationAngle);
-
-                qCDebug(qLcGstVideoRenderer) << "    sending video frame";
-                m_sink->setVideoFrame(frame);
-            }
-
-            gst_buffer_unref(buffer);
-
-            locker->relock();
-
-            m_renderReturn = GST_FLOW_OK;
-        }
-
-        m_renderCondition.wakeAll();
-    } else {
-        m_setupCondition.wakeAll();
-
-        return false;
-    }
-    return true;
+    // "data is to be discarded"
+    m_flushing = true;
+    m_bufferQueue.clear();
 }
 
-void QGstVideoRenderer::notify()
+void QGstVideoRenderer::gstEventHandleFlushStop(GstEvent *)
 {
-    if (!m_notified) {
-        m_notified = true;
-        QCoreApplication::postEvent(this, new QEvent(QEvent::UpdateRequest));
-    }
-}
-
-bool QGstVideoRenderer::waitForAsyncEvent(
-        QMutexLocker<QMutex> *locker, QWaitCondition *condition, unsigned long time)
-{
-    if (QThread::currentThread() == thread()) {
-        while (handleEvent(locker)) {}
-        m_notified = false;
-
-        return true;
-    }
-
-    notify();
-
-    return condition->wait(&m_mutex, time);
+    // "data is allowed again"
+    m_flushing = false;
 }
 
 static GstVideoSinkClass *gvrs_sink_parent_class;
@@ -407,15 +341,16 @@ static thread_local QGstreamerVideoSink *gvrs_current_sink;
 
 #define VO_SINK(s) QGstVideoRendererSink *sink(reinterpret_cast<QGstVideoRendererSink *>(s))
 
-QGstVideoRendererSink *QGstVideoRendererSink::createSink(QGstreamerVideoSink *sink)
+QGstVideoRendererSinkElement QGstVideoRendererSink::createSink(QGstreamerVideoSink *sink)
 {
     setSink(sink);
     QGstVideoRendererSink *gstSink = reinterpret_cast<QGstVideoRendererSink *>(
             g_object_new(QGstVideoRendererSink::get_type(), nullptr));
 
-    g_signal_connect(G_OBJECT(gstSink), "notify::show-preroll-frame", G_CALLBACK(handleShowPrerollChange), gstSink);
-
-    return gstSink;
+    return QGstVideoRendererSinkElement{
+        gstSink,
+        QGstElement::NeedsRef,
+    };
 }
 
 void QGstVideoRendererSink::setSink(QGstreamerVideoSink *sink)
@@ -439,16 +374,8 @@ GType QGstVideoRendererSink::get_type()
         nullptr                                            // value_table
     };
 
-    static const GType type = []() {
-        const auto result = g_type_register_static(
-                GST_TYPE_VIDEO_SINK, "QGstVideoRendererSink", &info, GTypeFlags(0));
-
-        // Register the sink type to be used in custom piplines.
-        // When surface is ready the sink can be used.
-        gst_element_register(nullptr, "qtvideosink", GST_RANK_PRIMARY, result);
-
-        return result;
-    }();
+    static const GType type = g_type_register_static(GST_TYPE_VIDEO_SINK, "QGstVideoRendererSink",
+                                                     &info, GTypeFlags(0));
 
     return type;
 }
@@ -518,42 +445,13 @@ void QGstVideoRendererSink::finalize(GObject *object)
     G_OBJECT_CLASS(gvrs_sink_parent_class)->finalize(object);
 }
 
-void QGstVideoRendererSink::handleShowPrerollChange(GObject *o, GParamSpec *p, gpointer d)
-{
-    Q_UNUSED(o);
-    Q_UNUSED(p);
-    QGstVideoRendererSink *sink = reinterpret_cast<QGstVideoRendererSink *>(d);
-
-    gboolean showPrerollFrame = true; // "show-preroll-frame" property is true by default
-    g_object_get(G_OBJECT(sink), "show-preroll-frame", &showPrerollFrame, nullptr);
-
-    if (!showPrerollFrame) {
-        GstState state = GST_STATE_VOID_PENDING;
-        GstClockTime timeout = 10000000; // 10 ms
-        gst_element_get_state(GST_ELEMENT(sink), &state, nullptr, timeout);
-        // show-preroll-frame being set to 'false' while in GST_STATE_PAUSED means
-        // the QMediaPlayer was stopped from the paused state.
-        // We need to flush the current frame.
-        if (state == GST_STATE_PAUSED)
-            sink->renderer->flush();
-    }
-}
-
 GstStateChangeReturn QGstVideoRendererSink::change_state(
         GstElement *element, GstStateChange transition)
 {
-    QGstVideoRendererSink *sink = reinterpret_cast<QGstVideoRendererSink *>(element);
-
-    gboolean showPrerollFrame = true; // "show-preroll-frame" property is true by default
-    g_object_get(G_OBJECT(element), "show-preroll-frame", &showPrerollFrame, nullptr);
-
-    // If show-preroll-frame is 'false' when transitioning from GST_STATE_PLAYING to
-    // GST_STATE_PAUSED, it means the QMediaPlayer was stopped.
-    // We need to flush the current frame.
-    if (transition == GST_STATE_CHANGE_PLAYING_TO_PAUSED && !showPrerollFrame)
-        sink->renderer->flush();
-
-    return GST_ELEMENT_CLASS(gvrs_sink_parent_class)->change_state(element, transition);
+    GstStateChangeReturn ret =
+            GST_ELEMENT_CLASS(gvrs_sink_parent_class)->change_state(element, transition);
+    qCDebug(qLcGstVideoRenderer) << "QGstVideoRenderer::change_state:" << transition << ret;
+    return ret;
 }
 
 GstCaps *QGstVideoRendererSink::get_caps(GstBaseSink *base, GstCaps *filter)
@@ -562,29 +460,24 @@ GstCaps *QGstVideoRendererSink::get_caps(GstBaseSink *base, GstCaps *filter)
 
     QGstCaps caps = sink->renderer->caps();
     if (filter)
-        caps = QGstCaps(gst_caps_intersect(caps.get(), filter), QGstCaps::HasRef);
+        caps = QGstCaps(gst_caps_intersect(caps.caps(), filter), QGstCaps::HasRef);
 
-    gst_caps_ref(caps.get());
-    return caps.get();
+    return caps.release();
 }
 
 gboolean QGstVideoRendererSink::set_caps(GstBaseSink *base, GstCaps *gcaps)
 {
     VO_SINK(base);
-
     auto caps = QGstCaps(gcaps, QGstCaps::NeedsRef);
 
-    qCDebug(qLcGstVideoRenderer) << "set_caps:" << caps.toString();
+    qCDebug(qLcGstVideoRenderer) << "set_caps:" << caps;
 
     if (caps.isNull()) {
         sink->renderer->stop();
-
         return TRUE;
-    } else if (sink->renderer->start(caps)) {
-        return TRUE;
-    } else {
-        return FALSE;
     }
+
+    return sink->renderer->start(caps);
 }
 
 gboolean QGstVideoRendererSink::propose_allocation(GstBaseSink *base, GstQuery *query)
@@ -629,6 +522,23 @@ gboolean QGstVideoRendererSink::event(GstBaseSink *base, GstEvent * event)
     return GST_BASE_SINK_CLASS(gvrs_sink_parent_class)->event(base, event);
 }
 
-QT_END_NAMESPACE
+QGstVideoRendererSinkElement::QGstVideoRendererSinkElement(QGstVideoRendererSink *element,
+                                                           RefMode mode)
+    : QGstBaseSink{
+          qGstCheckedCast<GstBaseSink>(element),
+          mode,
+      }
+{
+}
 
-#include "moc_qgstvideorenderersink_p.cpp"
+void QGstVideoRendererSinkElement::setActive(bool isActive)
+{
+    qGstVideoRendererSink()->renderer->setActive(isActive);
+}
+
+QGstVideoRendererSink *QGstVideoRendererSinkElement::qGstVideoRendererSink() const
+{
+    return reinterpret_cast<QGstVideoRendererSink *>(element());
+}
+
+QT_END_NAMESPACE
