@@ -30,9 +30,9 @@ class QGstreamerVideoDevices final : public QPlatformVideoDevices,
 {
 public:
     explicit QGstreamerVideoDevices(QPlatformMediaIntegration *integration);
-    ~QGstreamerVideoDevices();
+    ~QGstreamerVideoDevices() override;
 
-    QList<QCameraDevice> videoDevices() const override;
+    QList<QCameraDevice> videoInputs() const override;
     GstDevice *videoDevice(const QByteArray &id) const;
 
     void addDevice(QGstDeviceHandle);
